rree <- function(x, y, intercept=TRUE, gam, beta.init, type="LPRE", algorithm="MM", maxit=1000, cov.beta=TRUE, tol=sqrt(.Machine$double.eps)){
	n <- nrow(x)
	p <- ncol(x)
	if(intercept) xx <- cbind(1,x)
	if(!intercept) xx <- x
	p1 <- ncol(xx)
	if(missing(beta.init)) beta.init <- rep(0,p1)

	betaold <- beta.init
	lossvalue <- NULL

	if(class(x)!="matrix")	stop('"x" must be matrix.')
	if(class(y)!="numeric")	stop('"y" must be a numeric vector.')
	if(nrow(x) != length(y)) stop('The number of rows of "x" must be equal to the lentgh of "y".')
	if(gam<0) stop('"gam" must be a positive number.')
	if(class(beta.init)!="numeric")	stop('"beta.init" must be a numeric vector.')
	if(length(beta.init)!= p1)	stop(paste("The length of beta.init must be ", p1, ".", sep=""))
	if(type != "LPRE" && type != "LSRE")	stop('"type" must be "LPRE" or "LSRE".')
	if(algorithm != "MM" && algorithm != "BFGS" && algorithm != "hybrid")	stop('"algorithm" must be "MM", "BFGS", or "hybrid"')
	if(maxit<=0) stop('"maxit" must be a positive integer.')
	if(class(cov.beta)!="logical") stop('The class of "cov.beta" must be logical.')
	if(tol<=0) stop('"gam" must be a positive number.')

	if(gam>0){
		#B <- 1/2 * (diag(n) - 1/n)
		#xTBx <- t(xx) %*% B %*% xx
		xt1 <- crossprod(xx, rep(1,n))
		xTBx <- (t(xx) %*% xx - xt1 %*% t(xt1) / n) * 1/2
		if(algorithm=="MM"){
			result <- MM_Rcpp(xx, y, gam, beta.init, type, xTBx, maxit, tol)
			betanew <- c(result$beta)
			lossvalue <- c(result$loss)
			weight <- c(result$weight)
		}
		if(algorithm=="BFGS"){
			fit.bfgs <- optim(beta.init, loss_gam_Rcpp, gr=gr_loss_gam_Rcpp, xx=xx, y=y, gam=gam, type=type, method="BFGS")
			betanew <- c(fit.bfgs$par)
			lossvalue <- c(fit.bfgs$value)
			weight <- c(exp(-gam * loss.f(y,xx %*% betanew,type=type)))
			weight <- weight / sum(weight)
		}		
		if(algorithm=="hybrid"){
			result <- MM_Rcpp(xx, y, gam, beta.init, type, xTBx, maxit, tol)
			fit.bfgs <- optim(result$beta, loss_gam_Rcpp, gr=gr_loss_gam_Rcpp, xx=xx, y=y, gam=gam, type=type, method="BFGS")
			betanew <- c(fit.bfgs$par)
			lossvalue <- c(fit.bfgs$value)
			weight <- c(exp(-gam * loss.f(y,xx %*% betanew,type=type)))
			weight <- weight / sum(weight)
		}
	}
	if(gam==0){
		fit.gam0 <- optim(beta.init, loss_gam0_Rcpp, gr=gr_loss_gam0_Rcpp, xx=xx, y=y, type=type, method="BFGS")
		betanew <- c(fit.gam0$par)
		lossvalue <- c(fit.gam0$value)
		weight <- rep(1/n,n)
	}
	if(cov.beta && type=="LPRE"){
		cov.beta <- Cov_beta(gam,xx,betanew)
		ans <- list(beta=betanew, loss=lossvalue, cov.beta=cov.beta, weight=weight, intercept=intercept, x=x, y=y)
	}else{
		ans <- list(beta=betanew, loss=lossvalue, weight=weight, intercept=intercept, x=x, y=y)
	}
		class(ans) <- "rree"
		ans$call <- match.call()
	ans
}


loss.f <- function(y, xbeta, type){
	t0 <- exp(xbeta)
	if(type == "LPRE")  ans <- y / t0 + t0 / y + log(y)
	if(type == "LSRE")  ans <- (1 - y / t0)^2 + (1 - t0 / y)^2 + log(y)
	ans
}


C_gam_h <- function(gam){
	besselK(2+2*gam,gam) / (2^gam * besselK(2,0)^(1+gam))
}

C1_gam_h <- function(gam){
	-besselK(2+2*gam,gam) / ((1+gam) * 2^gam * besselK(2,0)^(1+gam))
}

C2_gam_h <- function(gam){
	ans0 <- 2^(1-2*gam) / ((2*gam+1)^2 * besselK(2,0)^(2*gam+1))
	ans1 <- gam * (2*gam+1) * besselK(4*gam+2, 2*gam-2)
	ans2 <- (1+gam+2*gam^2) * besselK(4*gam+2, 2*gam-1)
	ans0 * (ans1 + ans2)
}

Pihat_kngam <- function(k,gam,xx,beta){
	tmp <- exp(-gam * xx %*% beta)
	if(k==0) ans <- mean(tmp) 
	if(k==1) ans <- colMeans(matrix(tmp,nrow(xx),ncol(xx)) * xx)
	if(k==2){
		ans <- t(xx) %*% (matrix(tmp,nrow(xx),ncol(xx)) * xx)
		ans <- ans / nrow(xx)
	}
	ans
}

Delta_gam <- function(gam,xx,beta){
	ans1 <- C_gam_h(gam)^2 * C2_gam_h(gam) * Pihat_kngam(0,gam,xx,beta)^2 * Pihat_kngam(2,2*gam,xx,beta)
	ans2 <- (C_gam_h(gam)+C1_gam_h(gam))^2 * C_gam_h(2*gam) * Pihat_kngam(0,2*gam,xx,beta)^2 * tcrossprod(Pihat_kngam(1,gam,xx,beta))
	ans3 <- -2*C_gam_h(gam) * (C_gam_h(gam)+C1_gam_h(gam)) * (C_gam_h(2*gam)+C1_gam_h(2*gam)) * Pihat_kngam(0,gam,xx,beta) * tcrossprod(Pihat_kngam(1,2*gam,xx,beta), Pihat_kngam(1,gam,xx,beta))
	ans <- ans1 + ans2 + ans3
	ans
 }

J_gam <- function(gam,xx,beta){
	C_gam_h(gam) * C2_gam_h(gam/2) * Pihat_kngam(0,gam,xx,beta) * Pihat_kngam(2,gam,xx,beta) - (C_gam_h(gam)+C1_gam_h(gam))^2 * tcrossprod(Pihat_kngam(1,gam,xx,beta))
}

Cov_beta <- function(gam,xx,beta){
	inv_Jgam <- solve(J_gam(gam,xx,beta))
	ans <- inv_Jgam %*% Delta_gam(gam,xx,beta) %*% inv_Jgam	
	ans
}