// [[Rcpp::depends(RcppArmadillo)]]
#include <RcppArmadillo.h>
using namespace arma;  // use the Armadillo library for matrix computations
using namespace Rcpp;


// [[Rcpp::depends(RcppArmadillo)]]
double lossMMupdate_Rcpp(vec beta0, vec weight, vec piold, mat xx, vec y, mat xTBx, double gam, vec betaold, String type){
		vec t0 = exp(xx * beta0);
		vec t0y = t0/y;
		vec loss;
		if(type=="LPRE"){
			loss = t0y + 1/t0y + log(y);			
		}
		if(type=="LSRE"){
			vec one_n = ones(size(t0));
			loss = (one_n - t0y) % (one_n - t0y) + (one_n - 1/t0y) % (one_n - 1/t0y) + log(y);
		}
		double loss_1 = dot(weight, loss);
		double loss_2_1 = gam*gam/2 * dot(beta0, (xTBx * beta0)); 
		double loss_2_2 = -gam * dot(beta0, (xx.t() * piold + gam * xTBx * betaold));
		double loss_2 = (loss_2_1 + loss_2_2) / (1 + gam);
		return loss_1 + loss_2;
}

// [[Rcpp::depends(RcppArmadillo)]]
vec gr_lossMMupdate_Rcpp(vec beta0, vec weight, vec piold, mat xx, vec y, mat xTBx, double gam, vec betaold, String type){
		vec t0 = exp(xx * beta0);
		vec t0y = t0/y;
		vec gr_1_0;
		if(type=="LPRE"){
			gr_1_0 = weight % (t0y - 1/t0y);
		}
		if(type=="LSRE"){
			gr_1_0 = 2 * weight % (- t0y + t0y%t0y + 1/t0y - 1/(t0y%t0y));
		}
		vec gr_1 = xx.t() * gr_1_0;
		vec gr_2_1 = gam*gam * (xTBx * beta0); 
		vec gr_2_2 = -gam * (xx.t() * piold + gam * xTBx * betaold);
		vec gr_2 = (gr_2_1 + gr_2_2) / (1 + gam);
		return gr_1 + gr_2;
}


// [[Rcpp::depends(RcppArmadillo)]]
Rcpp::List optim_Rcpp(const vec& beta0, vec& weight, vec& piold, mat& xx, vec& y, mat& xTBx, double& gam, vec& betaold, String& type){

  Rcpp::Environment stats("package:stats"); 
  Rcpp::Function optim = stats["optim"];    

  Rcpp::List opt_results = optim(Rcpp::_["par"]    = beta0,
                                 Rcpp::_["fn"]     = Rcpp::InternalFunction(&lossMMupdate_Rcpp),
                                 Rcpp::_["gr"] = Rcpp::InternalFunction(&gr_lossMMupdate_Rcpp),
                                 Rcpp::_["method"] = "BFGS",
                                 Rcpp::_["weight"]     = weight,
                                 Rcpp::_["piold"]     = piold,
                                 Rcpp::_["xx"]     = xx,
                                 Rcpp::_["y"]     = y,
                                 Rcpp::_["xTBx"]     = xTBx,
                                 Rcpp::_["gam"]     = gam,
                                 Rcpp::_["betaold"]     = betaold,
                                 Rcpp::_["type"]     = type
                                 );

  // Extract and coerce from list.
  //arma::vec out = Rcpp::as<arma::vec>(opt_results[0]);

  return opt_results;
}

// [[Rcpp::depends(RcppArmadillo)]]
vec lossf_Rcpp(vec y, vec xbeta, String type){
	vec t0 = exp(xbeta);
	vec t0y = t0/y;
	vec ans;
	if(type == "LPRE"){
		ans = t0y + 1/t0y + log(y);
	}
	if(type == "LSRE"){
		vec one_n = ones(size(t0));
		ans = (one_n - t0y) % (one_n - t0y) + (one_n - 1/t0y) % (one_n - 1/t0y) + log(y);
	}
	return ans;
}


// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
double loss_gam0_Rcpp(vec beta0, mat xx, vec y, String type){
	vec xbeta = xx * beta0;
	vec t0 = exp(xbeta);
	vec t0y = t0/y;
	vec loss_all;
	if(type == "LPRE"){
		loss_all = t0y + 1/t0y + log(y);
	}
	if(type == "LSRE"){
		vec one_n = ones(size(t0));
		loss_all = (one_n - t0y) % (one_n - t0y) + (one_n - 1/t0y) % (one_n - 1/t0y) + log(y);
	}
	double ans = sum(loss_all);
	return ans;
}

// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
vec gr_loss_gam0_Rcpp(vec beta0, mat xx, vec y, String type){
	vec xbeta = xx * beta0;
	vec t0 = exp(xbeta);
	vec t0y = t0/y;
	vec gr_1;
	if(type=="LPRE"){
		gr_1 = t0y - 1/t0y;
	}
	if(type=="LSRE"){
		gr_1 = 2 * (- t0y + t0y%t0y + 1/t0y - 1/(t0y%t0y));
	}
	vec ans = xx.t() * gr_1;
	return ans;
}



// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
double loss_gam_Rcpp(vec beta0, mat xx, vec y, double gam, String type){
	vec xbeta = xx * beta0;
	vec t0 = exp(xbeta);
	vec t0y = t0/y;
	vec loss;
	if(type == "LPRE"){
		loss = t0y + 1/t0y + log(y);
	}
	if(type == "LSRE"){
		vec one_n = ones(size(t0));
		loss = (one_n - t0y) % (one_n - t0y) + (one_n - 1/t0y) % (one_n - 1/t0y) + log(y);
	}
	vec loss_i1 = exp(-gam * loss);
	vec loss_i2 = exp(-gam * xbeta);
	double ans = -1/gam * log(sum(loss_i1)) + 1/(1+gam) * log(sum(loss_i2));
	return ans;
}

// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
vec gr_loss_gam_Rcpp(vec beta0, mat xx, vec y, double gam, String type){
	vec xbeta = xx * beta0;
	vec t0 = exp(xbeta);
	vec t0y = t0/y;

	//first term
	vec loss_deno1_vec;
	vec gr_loss;
	if(type == "LPRE"){
		loss_deno1_vec = t0y + 1/t0y + log(y);
		gr_loss = t0y - 1/t0y;
	}
	if(type == "LSRE"){
		vec one_n = ones(size(t0));
		loss_deno1_vec = (one_n - t0y) % (one_n - t0y) + (one_n - 1/t0y) % (one_n - 1/t0y) + log(y);
		gr_loss = 2 * (- t0y + t0y%t0y + 1/t0y - 1/(t0y%t0y));
	}
	double loss_deno1 = sum(exp(-gam * loss_deno1_vec));
	vec loss1 = xx.t() * (exp(-gam * loss_deno1_vec) % gr_loss) / loss_deno1;

	//second term
	vec loss_deno2_vec = exp(-gam * xbeta);
	double loss_deno2 = sum(loss_deno2_vec);
	vec loss2 = xx.t() * loss_deno2_vec / loss_deno2;
	loss2 = -gam / (1 + gam) * loss2;

	//ans
	vec ans = loss1 + loss2;
	return ans;
}



// [[Rcpp::depends(RcppArmadillo)]]
List MM_update_Rcpp(mat xx, vec y, double gam, vec betaold, String type, mat xTBx){
	vec xbeta = xx * betaold;
	vec weight = exp(-gam * lossf_Rcpp(y,xbeta,type));
	weight = weight / sum(weight);
	vec theta = -gam * xbeta;
	vec piold = exp(theta) / sum(exp(theta));
	vec betatmp = betaold;
	List optimresult = optim_Rcpp(betatmp, weight, piold, xx, y, xTBx, gam, betaold, type);
	//vec betaout = Rcpp::as<arma::vec>(optimresult[0]);
	return optimresult;
}


// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::export]]
List MM_Rcpp(mat xx, vec y, double gam, vec betaold, String type, mat xTBx, double maxit, double tol){
	List MMupdateresult;
	vec betanew;
	vec betaold0 = betaold;
	vec lossvalue(maxit+1,fill::zeros);
	lossvalue(0) = loss_gam_Rcpp(betaold0, xx, y, gam, type);
	double th2 = 2*tol;
	int count = 0;
	for(int tt=0; tt < maxit; ++tt){
		MMupdateresult = MM_update_Rcpp(xx, y, gam, betaold0, type, xTBx);
		betanew = Rcpp::as<arma::vec>(MMupdateresult[0]);
		lossvalue(tt+1) = loss_gam_Rcpp(betanew, xx, y, gam, type);
		//th2 = fabs(lossvalue(tt+1) - lossvalue(tt)) / fabs(lossvalue(tt+1));
		th2 = sum((betanew - betaold0) % (betanew - betaold0)) / sum(betaold0 % betaold0);
		count = count + 1;
		if(tol > sqrt(th2)){
			break;
		}
		betaold0 = betanew;
    }

    int convergence = 1;
    if(count + 1 < maxit){
    	lossvalue.resize(count + 1);
    	convergence = 0;
    }


	vec xbeta = xx * betanew;
	vec weight = exp(-gam * lossf_Rcpp(y,xbeta,type));
	weight = weight / sum(weight);

    return Rcpp::List::create(Rcpp::Named("beta")=betanew,
    						  Rcpp::Named("loss")=lossvalue,
    						  Rcpp::Named("convergence")=convergence,
    						  Rcpp::Named("weight")=weight);
    
}


//END
